/*
 * linux/drivers/char/cs89712port.c
 *
 * Copyright (C) 2001 Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/hardware/clps7111.h>
#include <linux/cs89712port.h>

#define CHRDEV  "cs89712port"
#define VERSION CHRDEV ": CS89712 port driver, (C) 2001 Atmark Techno, Inc."

typedef struct {
  unsigned int minor;
  char         chrdev[6];
  u32          iobase;
  u8           value;
} cs89712port_t;

#define CS89712PORTTBL_NUM 8
static cs89712port_t cs89712portTbl[CS89712PORTTBL_NUM] = {
  {CS89712PORT_MINOR0, "padr" , CLPS7111_BASE + PADR , 0},
  {CS89712PORT_MINOR1, "pbdr" , CLPS7111_BASE + PBDR , 0},
  {CS89712PORT_MINOR2, "pddr" , CLPS7111_BASE + PDDR , 0},
  {CS89712PORT_MINOR3, "pedr" , CLPS7111_BASE + PEDR , 0},
  {CS89712PORT_MINOR4, "paddr", CLPS7111_BASE + PADDR, 0},
  {CS89712PORT_MINOR5, "pbddr", CLPS7111_BASE + PBDDR, 0},
  {CS89712PORT_MINOR6, "pdddr", CLPS7111_BASE + PDDDR, 0},
  {CS89712PORT_MINOR7, "peddr", CLPS7111_BASE + PEDDR, 0},
};

ssize_t cs89712port_get (unsigned int minor, char *buf)
{
  cs89712port_t *pTbl;
  
  switch (minor & 0xfffffff0) {
  case CS89712PORT_MINOR0:
    pTbl = cs89712portTbl;
    break;
  case CS89712PORT_MINOR1:
    pTbl = cs89712portTbl + 1;
    break;
  case CS89712PORT_MINOR2:
    pTbl = cs89712portTbl + 2;
    break;
  case CS89712PORT_MINOR3:
    pTbl = cs89712portTbl + 3;
    break;
  case CS89712PORT_MINOR4:
    pTbl = cs89712portTbl + 4;
    break;
  case CS89712PORT_MINOR5:
    pTbl = cs89712portTbl + 5;
    break;
  case CS89712PORT_MINOR6:
    pTbl = cs89712portTbl + 6;
    break;
  case CS89712PORT_MINOR7:
    pTbl = cs89712portTbl + 7;
    break;
  default:
    return -ENODEV;
  }
  
  minor &= 0xf;
  if (minor > 8) {
    return -ENODEV;
  }
  *buf = inb (pTbl->iobase);
  if (minor) {
    *buf = (*buf >> (minor - 1)) & 0x1;
  }
  
  return 0;
}

ssize_t cs89712port_put (unsigned int minor, const char *buf)
{
  cs89712port_t *pTbl;
  
  switch (minor & 0xfffffff0) {
  case CS89712PORT_MINOR0:
    pTbl = cs89712portTbl;
    break;
  case CS89712PORT_MINOR1:
    pTbl = cs89712portTbl + 1;
    break;
  case CS89712PORT_MINOR2:
    pTbl = cs89712portTbl + 2;
    break;
  case CS89712PORT_MINOR3:
    pTbl = cs89712portTbl + 3;
    break;
  case CS89712PORT_MINOR4:
    pTbl = cs89712portTbl + 4;
    break;
  case CS89712PORT_MINOR5:
    pTbl = cs89712portTbl + 5;
    break;
  case CS89712PORT_MINOR6:
    pTbl = cs89712portTbl + 6;
    break;
  case CS89712PORT_MINOR7:
    pTbl = cs89712portTbl + 7;
    break;
  default:
    return -ENODEV;
  }
  
  minor &= 0xf;
  if (minor > 8) {
    return -ENODEV;
  }
  if (minor) {
    if (*buf) {
      pTbl->value |= (0x1 << (minor - 1));
    } else {
      pTbl->value &= ~(0x1 << (minor - 1));
    }
  } else {
    pTbl->value = *buf;
  }
  outb (pTbl->value, pTbl->iobase);
  
  return 0;
}

EXPORT_SYMBOL(cs89712port_get);
EXPORT_SYMBOL(cs89712port_put);

static ssize_t cs89712port_read (struct file *file, char *buf, size_t count,
				 loff_t *ppos)
{
  char    tmp;
  ssize_t ret;
  
  ret = cs89712port_get (MINOR(file->f_dentry->d_inode->i_rdev), &tmp);
  if (ret) {
    return 1;
  }

  if (put_user (tmp, buf)) {
    return -EFAULT;
  }
  return 1;
}

static ssize_t cs89712port_write (struct file *file, const char *buf,
				  size_t count, loff_t *ppos)
{
  ssize_t ret;
  
  ret = cs89712port_put (MINOR(file->f_dentry->d_inode->i_rdev), buf);
  if (ret) {
    return ret;
  }
  
  return 1;
}

static struct file_operations cs89712port_fops = {
  owner: THIS_MODULE,
  read:  cs89712port_read,
  write: cs89712port_write,
};

static int __init cs89712port_init (void)
{
  int  i;
  
  for (i = 0; i < CS89712PORTTBL_NUM; i++) {
    if (128 <= cs89712portTbl[i].minor) {
      cs89712portTbl[i].value = inb (cs89712portTbl[i].iobase);
    }
    if (!request_region (cs89712portTbl[i].iobase, 1,
			cs89712portTbl[i].chrdev)) {
      printk (KERN_WARNING CHRDEV ": I/O port 0x%x is not free.\n",
	      cs89712portTbl[i].iobase);
      for (i--; i >=0; i--) {
	release_region (cs89712portTbl[i].iobase, 1);
      }
      return -EIO;
    }
  }
  
  if (register_chrdev (CS89712PORT_MAJOR, CHRDEV, &cs89712port_fops)) {
    printk (KERN_WARNING CHRDEV ": Unable to get major %d.\n",
	    CS89712PORT_MAJOR);
    for (i = CS89712PORTTBL_NUM; i >= 0; i--) {
      release_region (cs89712portTbl[i].iobase, 1);
    }
    return -EIO;
  }
  
  printk (KERN_INFO VERSION "\n");
  return 0;
}

static void __exit cs89712port_cleanup (void)
{
  int i;
  
  unregister_chrdev (CS89712PORT_MAJOR, CHRDEV);
  
  for (i = CS89712PORTTBL_NUM; i >= 0 ; i++) {
    release_region (cs89712portTbl[i].iobase, 1);
  }
}

module_init (cs89712port_init);
module_exit (cs89712port_cleanup);
