/*
 *  linux/drivers/i2c/i2c-armadillo.c
 *
 *  Copyright (C) 2001 Atmark Techno, Inc.
 */

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/miscdevice.h>
#include <linux/rtc.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/i2c-s3531a.h>
#include <linux/cs89712port.h>

#define CHRDEV  "i2c-armadillo"
#define VERSION CHRDEV ": i2c Armadillo driver, (C) 2001 Atmark Techno, Inc."

static struct i2c_client *rtc_client;

static inline int rtc_command(int cmd, void *data)
{
  int ret = -EIO;
  
  if (rtc_client) {
    ret = rtc_client->driver->command(rtc_client, cmd, data);
  }
  
  return ret;
}

int get_rtc_time(struct rtc_time *rtctm)
{
  struct s3531a_time rawtm;
  int ret;
  
  memset(rtctm, 0, sizeof(struct rtc_time));
  
  ret = rtc_command(RTC_GETDATETIME, &rawtm);
  if (!ret) {
    rtctm->tm_year = rawtm.year + ((rawtm.year <= 69) ? 100 : 0);
    rtctm->tm_mon  = rawtm.mon - 1;
    rtctm->tm_mday = rawtm.mday;
    rtctm->tm_wday = rawtm.wday;
    rtctm->tm_hour = rawtm.hour;
    rtctm->tm_min  = rawtm.min;
    rtctm->tm_sec  = rawtm.sec;
  }
  
  return ret;
}

static int set_rtc_time(struct rtc_time *rtctm)
{
  struct s3531a_time rawtm;
  
  rawtm.year = rtctm->tm_year - ((rtctm->tm_year < 100) ? 0 : 100);
  rawtm.mon  = rtctm->tm_mon + 1;
  rawtm.mday = rtctm->tm_mday;
  rawtm.wday = rtctm->tm_wday;
  rawtm.hour = rtctm->tm_hour;
  rawtm.min  = rtctm->tm_min;
  rawtm.sec  = rtctm->tm_sec;
  
  return rtc_command(RTC_SETDATETIME, &rawtm);
}

static int rtc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
  struct rtc_time rtctm;
  int ret;
  
  switch (cmd) {
  case RTC_ALM_READ:
    break;
    
  case RTC_ALM_SET:
    break;
    
  case RTC_RD_TIME:
    ret = get_rtc_time(&rtctm);
    if (ret) {
      return ret;
    }
    return copy_to_user((void *)arg, &rtctm, sizeof(rtctm)) ? -EFAULT : 0;
    
  case RTC_SET_TIME:
    if (copy_from_user(&rtctm, (struct rtc_time*)arg, sizeof(rtctm))) {
      return -EFAULT;
    }
    return set_rtc_time(&rtctm);
  }
  
  return -EINVAL;
}

static struct file_operations rtc_fops = {
  ioctl: rtc_ioctl,
};

static struct miscdevice rtc_dev = {
  minor: RTC_MINOR,
  name:	 "rtc",
  fops:	 &rtc_fops,
};

static void armadillo_setscl(void *data, int state)
{
  char buf;
  
  buf = 0;
  if (!state) {
    cs89712port_put (CS89712PORT_MINOR0 + 2, &buf);
    buf = 1;
  }
  cs89712port_put (CS89712PORT_MINOR4 + 2, &buf);
}

static void armadillo_setsda(void *data, int state)
{
  char buf;
  
  buf = 0;
  if (!state) {
    cs89712port_put (CS89712PORT_MINOR0 + 1, &buf);
    buf = 1;
  }
  cs89712port_put (CS89712PORT_MINOR4 + 1, &buf);
}

static int armadillo_getscl(void *data)
{
  char buf;
  
  if (0 > cs89712port_get (CS89712PORT_MINOR0 + 2, &buf)) {
    return -1;
  }
  
  return buf;
}

static int armadillo_getsda(void *data)
{
  char buf;
  
  if (0 > cs89712port_get (CS89712PORT_MINOR0 + 1, &buf)) {
    return -1;
  }
  
  return buf;
}

static struct i2c_algo_bit_data armadillo_data = {
  setsda:  armadillo_setsda,
  setscl:  armadillo_setscl,
  getsda:  armadillo_getsda,
  getscl:  armadillo_getscl,
  udelay:  5,
  mdelay:  5,
  timeout: 100
};

static int armadillo_client_reg(struct i2c_client *client)
{
  unsigned char ctrl;
  struct rtc_time first, now;
  
  if (client->id == I2C_DRIVERID_S3531A &&
      client->addr == I2C_S3531A_ADDRESS) {
    rtc_client = client;
    rtc_command(RTC_GETCTRL, &ctrl);
    if (!(ctrl & CTRL_12_24)) {
      ctrl |= CTRL_12_24;
      rtc_command(RTC_SETCTRL, &ctrl);
    }
    
    if (!get_rtc_time(&first)) {
      while (!get_rtc_time(&now)) {
	if (now.tm_sec != first.tm_sec) {
	  xtime.tv_usec = 0;
	  xtime.tv_sec = mktime(1900 + now.tm_year,
				now.tm_mon + 1, now.tm_mday,
				now.tm_hour, now.tm_min, now.tm_sec);
	  break;
	}
      }
    }
  }
  
  return 0;
}

static int armadillo_client_unreg(struct i2c_client *client)
{
  if (client == rtc_client) {
    rtc_client = NULL;
  }
  
  return 0;
}

static struct i2c_adapter armadillo_ops = {
  name:		     "ARMADILLO",
  id:		     I2C_HW_B_ARMADILLO,
  algo_data:	     &armadillo_data,
  client_register:   armadillo_client_reg,
  client_unregister: armadillo_client_unreg,
};

static int __init i2c_armadillo_init(void)
{
  char buf;
  int  ret;
  
  buf = 0;
  cs89712port_put (CS89712PORT_MINOR4 + 2, &buf);
  cs89712port_put (CS89712PORT_MINOR4 + 1, &buf);
  
  ret = i2c_bit_add_bus(&armadillo_ops);
  if (ret < 0) {
    printk (KERN_WARNING CHRDEV ": Unable to add %s bus.\n",
	    armadillo_ops.name);
    return ret;
  }
  
  if (misc_register(&rtc_dev)) {
    printk (KERN_WARNING CHRDEV ": Unable to register %s driver.\n",
	    rtc_dev.name);
    return ret;
  }
  
  printk (KERN_INFO VERSION "\n");
  return 0;
}

static void __exit i2c_armadillo_cleanup(void)
{
  misc_deregister(&rtc_dev);
  i2c_bit_del_bus(&armadillo_ops);
}

module_init(i2c_armadillo_init);
module_exit(i2c_armadillo_cleanup);
